/* 2011-12-15: File added by Sony Corporation */
/*
 * An I2C driver for the Philips RX8025 RTC
 * Copyright 2005-06 Tower Technologies
 *
 * Author: Alessandro Zummo <a.zummo@towertech.it>
 * Maintainers: http://www.nslu2-linux.org/
 *
 * based on the other drivers in this same directory.
 *
 * http://www.semiconductors.philips.com/acrobat/datasheets/RX8025-04.pdf
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/i2c.h>
#include <linux/bcd.h>
#include <linux/rtc.h>

#define DRV_VERSION "0.4.3"

#define RX8025_REG_SC  0x00		/* datetime */
#define RX8025_REG_MN  0x01
#define RX8025_REG_HR  0x02                          
#define RX8025_REG_DW  0x03
#define RX8025_REG_DM  0x04
#define RX8025_REG_MO  0x05
#define RX8025_REG_YR  0x06
                            
#define RX8025_REG_DO 0x07		/* digital offset */

#define RX8025_REG_AWM 0x08		/* alarm */
#define RX8025_REG_AWH  0x09    
#define RX8025_REG_AWW  0x0A 
#define RX8025_REG_ADM  0x0B
#define RX8025_REG_ADH  0x0C

#define RX8025_REG_ST1	0x0E	/* status */
#define RX8025_REG_ST2	0x0F
#define RX8025_SC_LV	0x40
#define ST1_REG_DEF		20


#define WARNING_LOW_VOLTAGE 1


static struct i2c_driver RX8025_driver;

struct RX8025 {
	struct rtc_device *rtc;
};



static int RX8025_i2c_write(struct i2c_client *client, unsigned char addr, unsigned char value){
	int result;
	unsigned char send_data[2];
	send_data[0] = (addr << 4) & 0xf0;
	send_data[1] = value;
	result = i2c_master_send(client, send_data, sizeof(send_data));
	
	dev_dbg(&client->dev, "RX8025 : write %02x %02x\n", addr, value);
	
	if(result != sizeof(send_data)){
		dev_dbg(&client->dev, "RX8025 : write error\n");
		return -1;
	}
	return 0;
}

static int RX8025_i2c_read(struct i2c_client *client, unsigned char addr, unsigned char *value){
	unsigned char addr_shift;
	struct i2c_msg msgs[] = {
		{ client->addr, 0, 1, &addr_shift },
		{ client->addr, I2C_M_RD, 1, value },
	};
	int result;
	
	addr_shift = (addr << 4) & 0xf0;
	result = i2c_transfer(client->adapter, msgs, ARRAY_SIZE(msgs));
	dev_dbg(&client->dev, "RX8025 : read %02x %02x\n", addr, *value);
	
	if(result != ARRAY_SIZE(msgs)){
		dev_dbg(&client->dev, "RX8025 : read error\n");
		return -1;
	}
	return 0;
}




/*
 * @brief  {RX8025時間、日付取得}RX8025時間を取得する。
 * @param[IN] <<client>> RX8025デバイス構造体
 * @param[OUT] <<tm>> 時間、日付の構造体
 * @return {状態}　状態を返す
 * @pre    なし
 * @post   なし
 * @attention なし
 */
static int RX8025_get_datetime(struct i2c_client *client, struct rtc_time *tm)
{
    int retval = 0;
	unsigned char buf[16] = { (RX8025_REG_SC << 4) & 0xf0 };
	struct i2c_msg msgs[] = {
		{ client->addr, 0, 1, buf },		/* setup read ptr */
		{ client->addr, I2C_M_RD, 16, buf },	/* read status + date */
	};

	/* read registers */
	if ((i2c_transfer(client->adapter, msgs, 2)) != 2) {
		dev_err(&client->dev, "%s: read error\n", __func__);
		return -EIO;
	}

	if (buf[RX8025_REG_ST2] & RX8025_SC_LV) {
		dev_info(&client->dev,
			"low voltage detected, date/time is not reliable.\n");
		retval = WARNING_LOW_VOLTAGE;
	}

	dev_dbg(&client->dev,
		"%s: raw data is st1=%02x, st2=%02x, sec=%02x, min=%02x, hr=%02x, "
		"mday=%02x, wday=%02x, mon=%02x, year=%02x\n",
		__func__,
		buf[14], buf[15], buf[0], buf[1],
		buf[2], buf[4], buf[3], buf[5],
		buf[6]);


	tm->tm_sec = bcd2bin(buf[RX8025_REG_SC] & 0x7F);
	tm->tm_min = bcd2bin(buf[RX8025_REG_MN] & 0x7F);
	tm->tm_hour = bcd2bin(buf[RX8025_REG_HR] & 0x3F); 	/* rtc hr 0-23 */
	tm->tm_mday = bcd2bin(buf[RX8025_REG_DM] & 0x3F);
	tm->tm_wday = buf[RX8025_REG_DW] & 0x07;
	tm->tm_mon = bcd2bin(buf[RX8025_REG_MO] & 0x1F) - 1; 	/* rtc mn 1-12 */
	tm->tm_year = bcd2bin(buf[RX8025_REG_YR] & 0xFF);
	tm->tm_year += 100;						/* assume we are in 2000...2099 */

	dev_dbg(&client->dev, "%s: tm is secs=%d, mins=%d, hours=%d, "
		"mday=%d, mon=%d, year=%d, wday=%d\n",
		__func__,
		tm->tm_sec, tm->tm_min, tm->tm_hour,
		tm->tm_mday, tm->tm_mon, tm->tm_year, tm->tm_wday);

	/* the clock can give out invalid datetime, but we cannot return
	 * -EINVAL otherwise hwclock will refuse to set the time on bootup.
	 */
	if ((rtc_valid_tm(tm) < 0) || (tm->tm_year < 100) || (tm->tm_year > 199))
		dev_err(&client->dev, "retrieved date/time is not valid.\n");

	return retval;
}

/*
 * @brief  {RX8025時間、日付設定}RX8025時間、日付を設定する。
 * @param[IN] <<client>> RX8025デバイス構造体
 * @param[OUT] <<tm>> 時間、日付の構造体
 * @return {状態}　状態を返す
 * @pre	   なし
 * @post   なし
 * @attention なし
 */
static int RX8025_set_datetime(struct i2c_client *client, struct rtc_time *tm)
{
	int i, err;
	unsigned char buf[9];
	unsigned char control_register2;
	int result;

	dev_dbg(&client->dev, "%s: secs=%d, mins=%d, hours=%d, "
		"mday=%d, mon=%d, year=%d, wday=%d\n",
		__func__,
		tm->tm_sec, tm->tm_min, tm->tm_hour,
		tm->tm_mday, tm->tm_mon, tm->tm_year, tm->tm_wday);

	if((tm->tm_year < 100) || (tm->tm_year > 199))
		return -EINVAL;
	/* hours, minutes and seconds */
	buf[RX8025_REG_SC] = bin2bcd(tm->tm_sec) & 0x7f;
	buf[RX8025_REG_MN] = bin2bcd(tm->tm_min) & 0x7f;
	buf[RX8025_REG_HR] = bin2bcd(tm->tm_hour) & 0x3f;
	buf[RX8025_REG_DM] = bin2bcd(tm->tm_mday) & 0x3f;

	/* month, 1 - 12 */
	buf[RX8025_REG_MO] = bin2bcd(tm->tm_mon + 1) & 0x1f;

	/* year and 12/24 hour switch */
	buf[RX8025_REG_YR] = bin2bcd(tm->tm_year % 100) & 0xff;

	buf[RX8025_REG_DW] = tm->tm_wday & 0x07;

	/* write register's data */
	for (i = 0; i < 7; i++) {
		unsigned char data[2] = { ((RX8025_REG_SC + i) << 4) & 0xf0,
						buf[RX8025_REG_SC + i] };

		err = i2c_master_send(client, data, sizeof(data));
		if (err != sizeof(data)) {
			dev_err(&client->dev,
				"%s: err=%d addr=%02x, data=%02x\n",
				__func__, err, data[0], data[1]);
			return -EIO;
		}
	}
	
	/* clear low voltage flag */
	result = RX8025_i2c_read(client, RX8025_REG_ST2, &control_register2);
	if(result != 0){
		dev_err(&client->dev, "%s: read error\n", __func__);
		return -EIO;
	}
	if( ((control_register2 >> 6) & 0x01) == 0x01){
		control_register2 &= ~(0x01 << 6);
		result = RX8025_i2c_write(client, RX8025_REG_ST2, control_register2);
		if(result != 0){
			dev_err(&client->dev, "%s: write error\n", __func__);
			return -EIO;
		}
	}

	return 0;
}


/*
 * @brief  {RTCレジスタ初期化}PowerOnResetが起こったときはRX8025のレジスタを初期化する。
 * @param[IN] <<client>> RX8025デバイス構造体
 * @return {状態} 状態を返す
 * @pre	   なし
 * @post   なし
 * @attention なし
 */
static int RX8025_reset_registers(struct i2c_client *client)
{
	int result;
	int i;
	unsigned char control_register2;
	
	result = RX8025_i2c_read(client, RX8025_REG_ST2, &control_register2);
	if(result != 0){
		return -ENODEV;
	}
	
	if( ( (control_register2 >> 4) & 0x01 ) == 0x01 ||		// PowerOnReset(bit4 = 1)
		( (control_register2 >> 5) & 0x01 ) == 0x00 ){		// OSC stopped (bit5 = 0)
		// reset RX8025 registers
		static const unsigned char init_value[][2] = {
			{RX8025_REG_ST1, 0x20}, // 24hour
			{RX8025_REG_ST2, 0xa0}, // OSC stop detection ON, voltage detection 1.3V
			
			{RX8025_REG_SC, 0x00},
			{RX8025_REG_MN, 0x00},
			{RX8025_REG_HR, 0x00},
			{RX8025_REG_DW, 0x00},
			{RX8025_REG_DM, 0x01},
			{RX8025_REG_MO, 0x01},
			{RX8025_REG_YR, 0x00},	// 2000/01/01 00:00
			
			{RX8025_REG_AWM, 0x00},
			{RX8025_REG_AWH, 0x00},
			{RX8025_REG_AWW, 0x00},
			{RX8025_REG_ADM, 0x00},
			{RX8025_REG_ADH, 0x00}, // Alarm OFF
		};
		
		for(i=0; i<ARRAY_SIZE(init_value); i++){
			result = RX8025_i2c_write(client, init_value[i][0], init_value[i][1]);
			if(result != 0){
				return -ENODEV;
			}
		}
	}
	
	return 0;
}



struct RX8025_limit
{
	unsigned char reg;
	unsigned char mask;
	unsigned char min;
	unsigned char max;
};

/*
 * @brief  {RTCデバイス検証}RTCデバイスはRX8025チップを検証する。
 * @param[IN] <<client>> RX8025デバイス構造体
 * @return {状態} 状態を返す
 * @pre    なし
 * @post   なし
 * @attention なし
 */
static int RX8025_validate_client(struct i2c_client *client)
{
	int i;

	static const struct RX8025_limit pattern[] = {
		/* register, mask, min, max */
		{ (RX8025_REG_SC << 4) & 0xf0,	0x7F,	0,	59	},
		{ (RX8025_REG_MN << 4) & 0xf0,	0x7F,	0,	59	},
		{ (RX8025_REG_HR << 4) & 0xf0,	0x3F,	0,	23	},
		{ (RX8025_REG_DM << 4) & 0xf0,	0x3F,	0,	31	},
		{ (RX8025_REG_MO << 4) & 0xf0,	0x1F,	0,	12	},
	};

	/* check limits (only registers with bcd values) */
	for (i = 0; i < ARRAY_SIZE(pattern); i++) {
		int xfer;
		unsigned char value;
		unsigned char buf = pattern[i].reg;

		struct i2c_msg msgs[] = {
			{ client->addr, 0, 1, &buf },
			{ client->addr, I2C_M_RD, 1, &buf },
		};

		xfer = i2c_transfer(client->adapter, msgs, ARRAY_SIZE(msgs));

		if (xfer != ARRAY_SIZE(msgs)) {
			dev_err(&client->dev,
				"%s: could not read register 0x%02X\n",
				__func__, pattern[i].reg);

			return -EIO;
		}

		value = bcd2bin(buf & pattern[i].mask);
		if (value > pattern[i].max ||
			value < pattern[i].min) {
			dev_dbg(&client->dev,
				"%s: pattern=%d, reg=%x, mask=0x%02x, min=%d, "
				"max=%d, value=%d, raw=0x%02X\n",
				__func__, i, pattern[i].reg, pattern[i].mask,
				pattern[i].min, pattern[i].max,
				value, buf);

			return -ENODEV;
		}
	}

	return 0;
}

/*
 * @brief  {RX8025時間、日付取得}RX8025時間を取得する。
 * @param[IN] <<client>> RX8025デバイス構造体
 * @param[OUT] <<tm>> 時間、日付の構造体
 * @return {状態} 状態を返す
 * @pre    なし
 * @post   なし
 * @attention なし
 */
static int RX8025_rtc_read_time(struct device *dev, struct rtc_time *tm)
{
	return RX8025_get_datetime(to_i2c_client(dev), tm);
}

/*
 * @brief  {RX8025時間、日付設定}RX8025時間、日付を設定する。
 * @param[IN] <<client>> RX8025デバイス構造体
 * @param[OUT] <<tm>> 時間、日付の構造体
 * @return {状態}　状態を返す
 * @pre    なし
 * @post   なし
 * @attention なし
 */
static int RX8025_rtc_set_time(struct device *dev, struct rtc_time *tm)
{
	return RX8025_set_datetime(to_i2c_client(dev), tm);
}

static const struct rtc_class_ops RX8025_rtc_ops = {
	.read_time	= RX8025_rtc_read_time,
	.set_time	= RX8025_rtc_set_time,
};

/*
 * @brief  {RX8025デバイス登録} RX8025デバイス登録を行う。
 * @param[IN] <<client>> RX8025デバイス構造体
 * @param[IN] <<id>> RX8025デバイス構造体のID
 * @return {状態} 状態を返す
 * @pre    なし
 * @post   なし
 * @attention なし
 */
static int RX8025_probe(struct i2c_client *client,
				const struct i2c_device_id *id)
{
	struct RX8025 *RX8025;
	unsigned char buf[1];
	unsigned char data[2];
	int err = 0;

	dev_dbg(&client->dev, "%s\n", __func__);

	if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C))
		return -ENODEV;

	RX8025 = kzalloc(sizeof(struct RX8025), GFP_KERNEL);
	if (!RX8025)
		return -ENOMEM;

	/* reset registers if PowerOnReset occurred */
	if (RX8025_reset_registers(client) < 0) {
		err = -ENODEV;
		goto exit_kfree;
	}
	
	/* Verify the chip is really an RX8025 */
	if (RX8025_validate_client(client) < 0) {
		err = -ENODEV;
		goto exit_kfree;
	}

	dev_info(&client->dev, "chip found, driver version " DRV_VERSION "\n");

	RX8025->rtc = rtc_device_register(RX8025_driver.driver.name,
				&client->dev, &RX8025_rtc_ops, THIS_MODULE);

	if (IS_ERR(RX8025->rtc)) {
		err = PTR_ERR(RX8025->rtc);
		goto exit_kfree;
	}

	i2c_set_clientdata(client, RX8025);
	buf[0] = bin2bcd(ST1_REG_DEF);
	data[0] = (RX8025_REG_ST1 << 4) & 0xf0;
	data[1] =  buf[0];
	err = i2c_master_send(client, data, sizeof(data));
	if (err != sizeof(data)) {	
		dev_err(&client->dev,"	%s: err=%d addr=%02x, data=%02x\n",
			__func__, err, data[0], data[1]);
		return -EIO;
	}

	return 0;

exit_kfree:
	kfree(RX8025);

	return err;
}

/*
 * @brief  { RX8025デバイス削除}カーネルからRX8025デバイスを削除する。
 * @param[IN] <<client>> RX8025デバイス構造体
 * @return {状態} 状態を返す
 * @pre    なし
 * @post   なし
 * @attention なし
 */
static int RX8025_remove(struct i2c_client *client)
{
	struct RX8025 *RX8025 = i2c_get_clientdata(client);

	if (RX8025->rtc)
		rtc_device_unregister(RX8025->rtc);

	kfree(RX8025);

	return 0;
}

static const struct i2c_device_id RX8025_id[] = {
	{ "rx8025", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, RX8025_id);

static struct i2c_driver RX8025_driver = {
	.driver		= {
		.name	= "rtc-RX8025",
	},
	.probe		= RX8025_probe,
	.remove	= RX8025_remove,
	.id_table	= RX8025_id,
};

/*
 * @brief  { RX8025デバイス登録}RX8025デバイスドライバをカーネルに登録する。
 * @param  なし
 * @return {状態} 状態を返す
 * @pre    なし
 * @post   なし
 * @attention なし
 */
static int __init RX8025_init(void)
{
	return i2c_add_driver(&RX8025_driver);
}

/*
 * @brief  {RX8025デバイスドライバ削除}カーネルからRX8025デバイスドライバを削除する。
 * @param なし
 * @return  なし
 * @pre    なし
 * @post   なし
 * @attention なし
 */
static void __exit RX8025_exit(void)
{
	i2c_del_driver(&RX8025_driver);
}

MODULE_AUTHOR("Alessandro Zummo <a.zummo@towertech.it>");
MODULE_DESCRIPTION("Epson RX8025 RTC driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);

module_init(RX8025_init);
module_exit(RX8025_exit);
